@@ -1,6 +1,6 @@
 #  [BSD 2-CLAUSE LICENSE]
 #
-#  Copyright Alex Yu 2021
+#  Copyright SVOX Authors 2021
 #
 #  Redistribution and use in source and binary forms, with or without
 #  modification, are permitted provided that the following conditions are met:
@@ -31,31 +31,28 @@ import os.path as osp
 import torch
 import numpy as np
 from torch import nn, autograd
-from svox.helpers import N3TreeView, _get_c_extension
+from svox.helpers import N3TreeView, DataFormat, _get_c_extension
 from warnings import warn
 
 _C = _get_c_extension()
 
 class _QueryVerticalFunction(autograd.Function):
     @staticmethod
-    def forward(ctx, data, child, indices, offset, invradius):
-        out = _C.query_vertical(data, child, indices, offset, invradius)
-        ctx.save_for_backward(child, indices, offset, invradius)
-        return out
+    def forward(ctx, data, tree_spec, indices):
+        out, node_ids = _C.query_vertical(tree_spec, indices)
+
+        ctx.mark_non_differentiable(node_ids)
+        ctx.tree_spec = tree_spec
+        ctx.save_for_backward(indices)
+        return out, node_ids
 
     @staticmethod
-    def backward(ctx, grad_out):
-        child, indices, offset, invradius = ctx.saved_tensors
-
-        grad_out = grad_out.contiguous()
+    def backward(ctx, grad_out, dummy):
         if ctx.needs_input_grad[0]:
-            grad_data = _C.query_vertical_backward(
-                    child, indices, grad_out,
-                    offset, invradius)
-        else:
-            grad_data = None
-
-        return grad_data, None, None, None, None
+            return _C.query_vertical_backward(ctx.tree_spec,
+                         ctx.saved_tensors[0],
+                         grad_out.contiguous()), None, None
+        return None, None, None
 
 
 class N3Tree(nn.Module):
@@ -65,14 +62,17 @@ class N3Tree(nn.Module):
     where :math:`N=2` is the familiar octree.
 
 .. warning::
-    `nn.Parameters` can change size, which 
+    `nn.Parameters` can change size, which
     makes current optimizers invalid. If any refine() or
     shrink_to_fit() call returns True,
     please re-make any optimizers
     """
     def __init__(self, N=2, data_dim=4, depth_limit=10,
             init_reserve=1, init_refine=0, geom_resize_fact=1.5,
-            radius=0.5, center=[0.5, 0.5, 0.5], map_location="cpu"):
+            radius=0.5, center=[0.5, 0.5, 0.5],
+            data_format="RGBA",
+            extra_data=None,
+            map_location="cpu"):
         """
         Construct N^3 Tree
 
@@ -82,8 +82,10 @@ class N3Tree(nn.Module):
         :param init_reserve: int amount of nodes to reserve initially
         :param init_refine: int number of times to refine entire tree initially
         :param geom_resize_fact: float geometric resizing factor
-        :param radius: float 1/2 side length of cube
+        :param radius: float or list, 1/2 side length of cube (possibly in each dim)
         :param center: list center of space
+        :param data_format: a string to indicate the data format
+        :param extra_data: extra data to include with tree
         :param map_location: str device to put data
 
         """
@@ -107,21 +109,28 @@ class N3Tree(nn.Module):
         self.register_buffer("_n_internal", torch.tensor(1, device=map_location))
         self.register_buffer("_n_free", torch.tensor(0, device=map_location))
 
-        radius = torch.tensor(radius, device=map_location)
-        center = torch.tensor(center, device=map_location)
-        
-        from packaging import version
-        if version.parse(torch.__version__) >= version.parse('1.6.0'):
-            self.register_buffer("invradius", 0.5 / radius, persistent=False)
-            self.register_buffer("offset", 0.5 * (1.0 - center / radius), persistent=False)
-        else:
-            self.register_buffer("invradius", 0.5 / radius)
-            self.register_buffer("offset", 0.5 * (1.0 - center / radius))
+        if isinstance(radius, float) or isinstance(radius, int):
+            radius = [radius] * 3
+        radius = torch.tensor(radius, dtype=torch.float32, device=map_location)
+        center = torch.tensor(center, dtype=torch.float32, device=map_location)
+
+        self.register_buffer("invradius", 0.5 / radius)
+        self.register_buffer("offset", 0.5 * (1.0 - center / radius))
 
         self.depth_limit = depth_limit
         self.geom_resize_fact = geom_resize_fact
+        self.data_format = DataFormat(data_format) if data_format is not None else None
+
+        if extra_data is not None:
+            assert isinstance(extra_data, torch.Tensor)
+            self.register_buffer("extra_data", extra_data.to(device=map_location))
+        else:
+            self.extra_data = None
+
         self._ver = 0
         self._invalidate()
+        self._lock_tree_structure = False
+        self._weight_accum = None
 
         self.refine(repeats=init_refine)
 
@@ -177,10 +186,7 @@ class N3Tree(nn.Module):
 
                 remain_mask &= nonterm_mask
         else:
-            _C.assign_vertical(self.data, self.child, indices,
-                               values,
-                               self.offset,
-                               self.invradius)
+            _C.assign_vertical(self._spec(), indices, values)
 
     def forward(self, indices, cuda=True, want_node_ids=False, world=True):
         """
@@ -198,7 +204,7 @@ class N3Tree(nn.Module):
         assert not indices.requires_grad  # Grad wrt indices not supported
         assert len(indices.shape) == 2
 
-        if not cuda or _C is None or not self.data.is_cuda or want_node_ids:
+        if not cuda or _C is None or not self.data.is_cuda:
             if not want_node_ids:
                 warn("Using slow query")
             if world:
@@ -244,11 +250,9 @@ class N3Tree(nn.Module):
 
             return result
         else:
-            return _QueryVerticalFunction.apply(
-                    self.data, self.child, indices,
-                    self.offset if world else torch.tensor(
-                        [0.0, 0.0, 0.0], device=indices.device),
-                    self.invradius if world else 1.0)
+            result, node_ids = _QueryVerticalFunction.apply(
+                                self.data, self._spec(world), indices);
+            return (result, node_ids.long()) if want_node_ids else result
 
     # Special features
     def snap(self, indices):
@@ -266,16 +270,17 @@ class N3Tree(nn.Module):
         """
         Get partial tree with some of the data dimensions (channels)
         E.g. tree.partial(-1) to get tree with data_dim 1 of last channel only
-
         :param data_sel: data channel selector, default is all channels
-
         :return: partial N3Tree (copy)
-
         """
-        sel_indices = torch.arange(self.data_dim)[data_sel]
-        if sel_indices.ndim == 0:
-            sel_indices = sel_indices.unsqueeze(0)
-        t2 = N3Tree(N=self.N, data_dim=sel_indices.numel(),
+        if data_sel is None:
+            new_data_dim = self.data_dim
+        else:
+            sel_indices = torch.arange(self.data_dim)[data_sel]
+            if sel_indices.ndim == 0:
+                sel_indices = sel_indices.unsqueeze(0)
+            new_data_dim = sel_indices.numel()
+        t2 = N3Tree(N=self.N, data_dim=new_data_dim,
                 depth_limit=self.depth_limit,
                 geom_resize_fact=self.geom_resize_fact,
                 map_location=self.data.data.device)
@@ -284,14 +289,25 @@ class N3Tree(nn.Module):
         t2.child = self.child.clone()
         t2.parent_depth = self.parent_depth.clone()
         t2._n_internal = self._n_internal.clone()
-        t2.data.data = self.data.data[..., sel_indices].contiguous()
+        t2._n_free = self._n_free.clone()
+        if data_sel is None:
+            t2.data.data = self.data.data.clone()
+        else:
+            t2.data.data = self.data.data[..., sel_indices].contiguous()
         return t2
 
+    def clone(self):
+        """
+        Deep copy the tree
+        """
+        return self.partial()
+
     # 'Frontier' operations (node merging/pruning)
     def merge(self, frontier_sel=None, op=torch.mean):
         """
         Merge leaves into selected 'frontier' nodes
         (i.e., nodes for which all children are leaves).
+        Use shrink_to_fit() to recover memory freed.
 
         :param frontier_sel: selector (int, mask, list of indices etc)
                              for frontier nodes. In same order as reduce_frontier().
@@ -300,12 +316,14 @@ class N3Tree(nn.Module):
                              conditions for merge, then pass
                              mask or indices to merge().
         :param op: reduction to combine child leaves into node.
-                   E.g. torch.max, torch.mean. 
+                   E.g. torch.max, torch.mean.
                    Should take a positional argument :code:`x` (B, N, data_dim) and
                    a named parameter :code:`dim` (always 1),
                    and return a matrix of (B, your_out_dim).
                    If a tuple is returned, uses first result.
         """
+        if self.n_internal - self._n_free.item() <= 1:
+            raise RuntimeError("Cannot merge root node")
         nid = self._frontier[frontier_sel]
         if nid.numel() == 0:
             return False
@@ -330,7 +348,7 @@ class N3Tree(nn.Module):
         (i.e., nodes for which all children are leaves).
 
         :param op: reduction to combine child leaves into node.
-                   E.g. torch.max, torch.mean. 
+                   E.g. torch.max, torch.mean.
                    Should take a positional argument :code:`x`
                    (B, N, in_dim <= data_dim) and
                    a named parameter :code:`dim` (always 1),
@@ -361,12 +379,6 @@ class N3Tree(nn.Module):
         operation, taking the returned values and discarding the
         argmax part.
 
-        :param op: reduction to combine child leaves into node.
-                   E.g. torch.max, torch.mean. 
-                   Should take a positional argument :code:`x`
-                   (B, N, in_dim <= data_dim) and
-                   a named parameter :code:`dim` (always 1),
-                   and return a matrix of (B, your_out_dim).
         :param dim: dimension(s) of data to return, e.g. -1 returns
                     last data dimension for all 'frontier' nodes
         :param grad: if True, returns a tensor differentiable wrt tree data.
@@ -377,6 +389,39 @@ class N3Tree(nn.Module):
         return self.reduce_frontier(op=lambda x, dim: torch.max(x, dim=dim)[0],
                 grad=grad, dim=dim)
 
+    def diam_frontier(self, dim=None, grad=False, scale=1.0):
+        """
+        Takes diameter over child leaf values for each 'frontier' node
+        (i.e., nodes for which all children are leaves).
+
+        :param dim: dimension(s) of data to return, e.g. -1 returns
+                    last data dimension for all 'frontier' nodes
+        :param grad: if True, returns a tensor differentiable wrt tree data.
+                      Default False.
+
+        :return: reduced tensor
+        """
+        def diam_func(x, dim):
+            # (B, N3, in_dim)
+            if x.ndim == 2:
+                x = x[:, :, None]
+            N3 = x.shape[1]
+            diam = torch.zeros(x.shape[:-2], device=x.device)
+            for offset in range(N3):
+                end_idx = -offset if offset > 0 else N3
+                delta = (x[:, offset:] - x[:, :end_idx]) * scale
+                n1 = torch.norm(delta, dim=-1).max(dim=-1)[0]
+                if offset:
+                    delta = (x[:, :offset] - x[:, end_idx:]) * scale
+                    n2 = torch.norm(delta, dim=-1).max(dim=-1)[0]
+                    n1 = torch.max(n1, n2)
+                diam = torch.max(diam, n1)
+            return diam
+
+        return self.reduce_frontier(op=diam_func,
+                grad=grad, dim=dim)
+
+
     @property
     def _frontier(self):
         """
@@ -393,12 +438,11 @@ class N3Tree(nn.Module):
 
 
     # Leaf refinement & memory management methods
-    def refine(self, repeats=1, dim=-1, thresh=None, max_refine=None, sel=None):
+    def refine(self, repeats=1, sel=None):
         """
-        Refine each leaf node, optionally filtering by value at dimension 'dim' >= 'thresh'.
-        Respects depth_limit.
+        Refine each selected leaf node, respecting depth_limit.
 
-        :param repeats: int number of times to repeat
+        :param repeats: int number of times to repeat refinement
         :param sel: (N, 4) node selector. Default selects all leaves.
 
         :return: True iff N3Tree.data parameter was resized, requiring
@@ -411,8 +455,10 @@ class N3Tree(nn.Module):
 .. warning::
     The selector :code:`sel` is assumed to contain unique leaf indices. If there are duplicates
     memory will be wasted. We do not dedup here for efficiency reasons.
-    
+
         """
+        if self._lock_tree_structure:
+            raise RuntimeError("Tree locked")
         with torch.no_grad():
             resized = False
             for repeat_id in range(repeats):
@@ -472,6 +518,8 @@ class N3Tree(nn.Module):
                     in xyz orde rto identify leaf within internal node
 
         """
+        if self._lock_tree_structure:
+            raise RuntimeError("Tree locked")
         assert min(xyzi) >= 0 and max(xyzi) < self.N
         if self.parent_depth[intnode_idx, 1] >= self.depth_limit:
             return
@@ -503,10 +551,13 @@ class N3Tree(nn.Module):
         """
         Shrink data & buffers to tightly needed fit tree data,
         possibly dealing with fragmentation caused by merging.
+        This is called by the save() function.
 
 .. warning::
         Will change the nn.Parameter size (data), breaking optimizer!
         """
+        if self._lock_tree_structure:
+            raise RuntimeError("Tree locked")
         n_int = self.n_internal
         n_free = self._n_free.item()
         new_cap = n_int - n_free
@@ -530,6 +581,8 @@ class N3Tree(nn.Module):
             self.data = nn.Parameter(self.data.data[remain_ids])
             self.child = self.child[remain_ids]
             self.parent_depth = self.parent_depth[remain_ids]
+            self._n_internal.fill_(new_cap)
+            self._n_free.zero_()
         else:
             # Direct resize
             self.data = nn.Parameter(self.data.data[:new_cap])
@@ -553,15 +606,33 @@ class N3Tree(nn.Module):
 
     @property
     def max_depth(self):
+        """
+        Maximum tree depth - 1
+        """
         return torch.max(self.depths).item()
 
+    def accumulate_weights(self):
+        """
+        Begin weight accumulation
+
+.. code-block:: python
+
+        with tree.accumulate_weights() as accum:
+            ...
+
+        # (n_leaves) in same order as values etc.
+        accum = accum()
+        """
+        return WeightAccumulator(self)
+
     # Persistence
-    def save(self, path, shrink=True):
+    def save(self, path, shrink=True, compress=True):
         """
         Save to from npz file
 
         :param path: npz path
-        :pram shrink: if True (default), applies shrink_to_fit before saving
+        :param shrink: if True (default), applies shrink_to_fit before saving
+        :param compress: whether to compress the npz; may be slow
 
         """
         if shrink:
@@ -572,13 +643,20 @@ class N3Tree(nn.Module):
             "parent_depth" : self.parent_depth.cpu(),
             "n_internal" : self._n_internal.cpu().item(),
             "n_free" : self._n_free.cpu().item(),
-            "invradius" : self.invradius.cpu().item(),
+            "invradius3" : self.invradius.cpu(),
             "offset" : self.offset.cpu(),
             "depth_limit": self.depth_limit,
             "geom_resize_fact": self.geom_resize_fact,
             "data": self.data.data.cpu().numpy().astype(np.float16)
         }
-        np.savez_compressed(path, **data)
+        if self.data_format is not None:
+            data["data_format"] = repr(self.data_format)
+        if self.extra_data is not None:
+            data["extra_data"] = self.extra_data.cpu()
+        if compress:
+            np.savez_compressed(path, **data)
+        else:
+            np.savez(path, **data)
 
     @classmethod
     def load(cls, path, map_location='cpu'):
@@ -596,23 +674,31 @@ class N3Tree(nn.Module):
         tree.N = tree.child.shape[-1]
         tree.parent_depth = torch.from_numpy(z["parent_depth"]).to(map_location)
         tree._n_internal.fill_(z["n_internal"].item())
-        tree.invradius.fill_(z["invradius"].item())
+        if "invradius3" in z.files:
+            tree.invradius = torch.from_numpy(z["invradius3"].astype(
+                                np.float32)).to(map_location)
+        else:
+            tree.invradius.fill_(z["invradius"].item())
         tree.offset = torch.from_numpy(z["offset"].astype(np.float32)).to(map_location)
         tree.depth_limit = int(z["depth_limit"])
         tree.geom_resize_fact = float(z["geom_resize_fact"])
         tree.data.data = torch.from_numpy(z["data"].astype(np.float32)).to(map_location)
-        if 'n_free' in z.keys():
+        if 'n_free' in z.files:
             tree._n_free.fill_(z["n_free"].item())
         else:
             tree._n_free.zero_()
+        tree.data_format = DataFormat(z['data_format'].item()) if \
+                'data_format' in z.files else None
+        tree.extra_data = torch.from_numpy(z['extra_data']).to(map_location) if \
+                          'extra_data' in z.files else None
         return tree
 
     # Magic
     def __repr__(self):
-        return ("svox.N3Tree(N={}, data_dim={}, depth_limit={};" +
-                " capacity:{}/{})").format(
-                    self.N, self.data_dim, self.depth_limit,
-                    self.n_internal - self._n_free.item(), self.capacity)
+        return (f"svox.N3Tree(N={self.N}, data_dim={self.data_dim}, " +
+                f"depth_limit={self.depth_limit}, " +
+                f"capacity:{self.n_internal - self._n_free.item()}/{self.capacity}, " +
+                f"data_format:{self.data_format or 'RGBA'})");
 
     def __getitem__(self, key):
         """
@@ -655,7 +741,7 @@ class N3Tree(nn.Module):
             good_mask = curr[:, 0] != 0
             if not good_mask.any():
                 break
-            mask[mask] = good_mask
+            mask[mask.clone()] = good_mask
 
             curr = self._unpack_index(self.parent_depth[curr[good_mask, 0], 0].long())
 
@@ -724,13 +810,32 @@ class N3Tree(nn.Module):
         self._last_all_leaves = None
         self._last_frontier = None
 
+    def _spec(self, world=True):
+        """
+        Pack tree into a TreeSpec (for passing data to C++ extension)
+        """
+        tree_spec = _C.TreeSpec()
+        tree_spec.data = self.data
+        tree_spec.child = self.child
+        tree_spec.extra_data = self.extra_data if self.extra_data is not None else \
+                torch.empty((0, 0), device=self.data.device)
+        tree_spec.offset = self.offset if world else torch.tensor(
+                  [0.0, 0.0, 0.0], device=self.data.device)
+        tree_spec.scaling = self.invradius if world else torch.tensor(
+                  [1.0, 1.0, 1.0], device=self.data.device)
+        if hasattr(self, '_weight_accum'):
+            tree_spec._weight_accum = self._weight_accum if \
+                    self._weight_accum is not None else torch.empty(
+                            0, device=self.data.device)
+        return tree_spec
+
 
 # Redirect functions to N3TreeView so you can do tree.depths instead of tree[:].depths
 def _redirect_to_n3view():
     redir_props = ['depths', 'lengths', 'lengths_local', 'corners', 'corners_local',
                    'values', 'values_local', 'ndim', 'shape']
-    redir_funcs = ['sample', 'sample_local', 'dim', 'numel', 'size', '__len__',
-            'normal_', 'clamp_', 'uniform_', 'relu_', 'sigmoid_']
+    redir_funcs = ['sample', 'sample_local', 'aux', 'dim', 'numel', 'size', '__len__',
+            'normal_', 'clamp_', 'uniform_', 'relu_', 'sigmoid_', 'nan_to_num_']
     def redirect_func(redir_func):
         def redir_impl(self, *args, **kwargs):
             return getattr(self[:], redir_func)(*args, **kwargs)
@@ -744,3 +849,26 @@ def _redirect_to_n3view():
     for redir_prop in redir_props:
         redirect_prop(redir_prop)
 _redirect_to_n3view()
+
+class WeightAccumulator():
+    def __init__(self, tree):
+        self.tree = tree
+
+    def __enter__(self):
+        self.tree._lock_tree_structure = True
+        self.tree._weight_accum = torch.zeros(
+                self.tree.child.shape, dtype=torch.float32,
+                device=self.tree.data.device)
+        self.weight_accum = self.tree._weight_accum
+        return self
+
+    def __exit__(self, type, value, traceback):
+        self.tree._weight_accum = None
+        self.tree._lock_tree_structure = False
+
+    @property
+    def value(self):
+        return self.weight_accum
+
+    def __call__(self):
+        return self.tree.aux(self.weight_accum)
