@@ -28,11 +28,11 @@ def xyz_to_int(q):
 for i in tqdm(range(5000)):
     if np.random.rand() < 0.5:
         q = torch.rand((1, 3), device=device)
-        vals = 100 * torch.randn((1, K), device=device)
+        vals = 100 * torch.randn((1, K), dtype=torch.float32, device=device)
 
         x, y, z = xyz_to_int(q)
         gt[x, y, z] = vals[0]
-        g.set(q, vals)
+        g[q] = vals
 
     # ---
 
@@ -40,13 +40,13 @@ for i in tqdm(range(5000)):
     x, y, z = xyz_to_int(q)
     rt = gt[x, y, z]
 
-    r = g(q, cuda=True)[0]
+    r = g[q].values[0]
     err = torch.abs(r - rt).max()
-    if err > 5e-5:
+    worst = max(err, worst)
+    if err > 5e-2:
         print(i, err.item(), 'q', q[0].detach().cpu().numpy(),
                 'g', r.detach().cpu().numpy(), 'gt', rt.detach().cpu().numpy(),
                  x, y, z)
         break
-    worst = max(err, worst)
 print('worst', worst)
 
