@@ -4,20 +4,21 @@ import svox
 from tqdm import tqdm
 
 device = 'cuda:0'
-torch.random.manual_seed(1234)
 torch.cuda.set_device(device)
+#  device = 'cpu'
+torch.random.manual_seed(1234)
 N = 64
 K = 4
 g = svox.N3Tree(N=K, map_location=device).to(device=device)
 
 q = torch.rand((1, 3), device=device)
 print('begin refine')
-for j in tqdm(range(1000)):
-    for i in range(10):
-        g._refine_at(j, torch.randint(0, K, (3,)))
+for j in tqdm(range(10)):
+    q = torch.rand((100, 3), device=device)
+    g[q].refine()
     for i in range(20):
         q = torch.rand((3, 3), device=device)
-        vals = 100 * torch.randn((1, K), device=device)
+        vals = 100 * torch.randn((3, K), device=device)
         g.set(q, vals)
 #
 q = torch.rand((100000, 3), device=device)
